################################################################################
###############              Quasi Sequence Order                ###############
################################################################################

import os
import json
import pandas as pd
import math
from json import JSONDecodeError
import sys
from . import composition

#list of amino acids
amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
    "Q", "R", "S", "T", "V", "W", "Y"]

def seq_order_coupling_number(sequence, lag=30,
    distance_matrix="schneider-wrede-physiochemical-distance-matrix.json"):
    """
    Calculate Sequence Order Coupling Number (SOCNum) features for the protein sequences.
    Sequence Order Coupling Number computes the dissimilarity between amino acid
    pairs. The distance between amino acid pairs is determined by d which varies
    between 1 to nlag. For each d, it computes the sum of the dissimilarities
    of all amino acid pairs. The number of output features can be calculated as N * 2,
    where N = lag, by default this value is 30 so 60 features are output.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.
    :lag : int (default=30)
        lag is the maximum lag; the length of the protein should be larger
        than lag. Default set to 30.
    :distance_matrix : str (default="schneider-wrede-physicochemical-distance-matrix")
        path to physiochemical distance matrix for calculating quasi sequence order.

    Returns
    -------
    :seq_order_df : pd.Dataframe
        Dataframe of SOCNum descriptor values for all protein sequences. Output
        will be of the shape N x 1, where N is the number of features calculated 
        from the descriptor (calculated as M * 2 where M = lag).

    References
    ----------
    [1]: Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating
        Quasi-Sequence-Order Effect. Biochemical and Biophysical Research Communications,
        2000, 278, 477-483.
    [2]: Kuo-Chen Chou and Yu-Dong Cai. Prediction of Protein Sucellular Locations by
        GO-FunD-PseAA Predictor. Biochemical and Biophysical Research Communications,
        2004, 320, 1236-1239.
    [3]: Gisbert Schneider and Paul Wrede. The Rational Design of Amino Acid Sequences
        by Artifical Neural Networks and Simulated Molecular Evolution: Do Novo Design
        of an Idealized Leader Cleavge Site. Biophys Journal, 1994, 66, 335-344.
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    #append extension if missing from input
    if (os.path.splitext(distance_matrix)[1] == ''):
        distance_matrix = distance_matrix + '.json'

    #get filepath to distance matrix json
    if not (os.path.isfile(distance_matrix)):
        if not (os.path.isfile(os.path.join(os.path.dirname(os.path.abspath(sys.modules['protpy'].__file__)), "data", distance_matrix))):
            raise OSError('Distance Matrix json ({}) not found.'.format(os.path.join(os.path.dirname(os.path.abspath(sys.modules['protpy'].__file__)), "data", distance_matrix)))
        else:
            distance_matrix = os.path.join(os.path.dirname(os.path.abspath(sys.modules['protpy'].__file__)), "data", distance_matrix)
    else:
        distance_matrix = distance_matrix
    
    #open distance matrix json if present
    try:
        with open(distance_matrix, "r") as f:
            distance_matrix = json.load(f)
    except:
        raise JSONDecodeError('Error getting config JSON file: {}.'.format(distance_matrix))

    #set default lag if invalid value input
    if (lag>=len(sequence) or (lag<0) or not (isinstance(lag, int))):
        lag=30

    seq_order = {}

    #iterate through sequence, calculating the SOCNum using the selected distance matrix
    for i in range(lag):
        tau = 0.0
        for j in range(len(sequence) - (i+1) ):
            current_aa = sequence[j]
            next_aa = sequence[j + (i+1)]
            tau = round(tau + math.pow(distance_matrix[current_aa + next_aa], 2),3)

        #append SOCNum of current lag to seq_order dict
        seq_order["SOCN" + str(i + 1)] = tau

    #transform descriptor data into pandas dataframe
    seq_order_df = pd.DataFrame([list(seq_order.values())], columns=list(seq_order.keys()))

    return seq_order_df

def quasi_sequence_order(sequence, max_lag=30, weight=0.1,
    distance_matrix="schneider-wrede-physicochemical-distance-matrix.json"):
    """
    Calculate Quasi Sequence Order features for the protein sequences.
    The quasi-sequence-order descriptors were proposed by K.C. Chou, et.al. [1].
    They are derived from the distance matrix between the 20 amino acids. By default,
    the Scheider-Wrede physicochemical distance matrix was used. Also utilised in
    the descriptor calculation is the Grantham chemical distance matrix. Both of
    these matrices are used by Grantham et. al. in the calculation
    of the descriptor [4]. 100 values are calculated per sequence, thus generating
    an output of 100 x 1 per sequence.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.
    :max_lag : int (default=30)
        A value for a lag, the max value is equal to the length of shortest peptide minus one.
    :weight: float (default = 0.1)
        weighting factor
    :distance_matrix : str (default="schneider-wrede-physicochemical-distance-matrix")
        path to physiochemical distance matrix for calculating quasi sequence order.

    Returns
    -------
    :quasi_seq_order_df : pd.Dataframe
        dataframe of quasi-sequence-order descriptor values for the
        protein sequences, with output shape 100 x 1 where 100 is the 
        number of calculated features per sequence.

    References
    ----------
    [1]: Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating
        Quasi-Sequence-Order Effect. Biochemical and Biophysical Research Communications
        2000, 278, 477-483.
    [2]: Kuo-Chen Chou and Yu-Dong Cai. Prediction of Protein sucellular locations by
        GO-FunD-PseAA predictor, Biochemical and Biophysical Research Communications,
        2004, 320, 1236-1239.
    [3]: Gisbert Schneider and Paul wrede. The Rational Design of Amino Acid
        Sequences by Artifical Neural Networks and Simulated Molecular Evolution: Do
        Novo Design of an Idealized Leader Cleavge Site. Biophys Journal, 1994, 66, 335-344.
    [4]: Grantham, R. (1974-09-06). "Amino acid difference formula to help explain protein
        evolution". Science. 185 (4154): 862–864. Bibcode:1974Sci...185..862G.
        doi:10.1126/science.185.4154.862. ISSN 0036-8075. PMID 4843792. S2CID 35388307.
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))
            
    #append extension if missing from input
    if (os.path.splitext(distance_matrix)[1] == ''):
        distance_matrix = distance_matrix + '.json'

    quasi_seq_order = {}
    right_part = 0.0

    #calculate quasi sequence order using sequence order coupling number for
    #proteins using lag and specificed physiochemical distance matrix
    for i in range(max_lag):
        right_part = right_part + seq_order_coupling_number(
            sequence, i + 1, distance_matrix
        )

    aa_comp = composition.AAComposition(sequence)
    temp = 1 + weight * right_part
    
    for index, i in enumerate(amino_acids):
        quasi_seq_order["Quasi_seq_order1_" + str(index + 1)] = round(aa_comp[i] / temp, 6)

    right_part = []
    for i in range(max_lag):
        right_part.append(
            seq_order_coupling_number(sequence, i+1, distance_matrix)
            )

    temp = 1 + weight * sum(right_part)
    for index in range(20, 20 + max_lag):
        quasi_seq_order["Quasi_seq_order2_" + str(index + 1)] = round(
            weight * right_part[index - 20] / temp, 6
        )

    #transform descriptor data into pandas dataframe
    quasi_seq_order_df = pd.DataFrame([list(quasi_seq_order.values())], columns=list(quasi_seq_order.keys()))

    return quasi_seq_order_df


###########################################################################
#############      protPy - Autocorrelation Module Tests      #############
###########################################################################

import pandas as pd
import numpy as np
import os
import unittest
import re
from Bio import SeqIO
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import protpy as protpy

class ProtPyAutocorrelationTests(unittest.TestCase):
    """
    Test suite for testing autcorrelation module and functionality 
    in protpy package. 

    Test Cases
    ----------
    test_moreaubroto_autocorrelation:
        testing correct protpy MoreauBroto Autocorrelation functionality.
    test_moran_autocorrelation:
        testing correct protpy Moran Autocorrelation functionality.
    test_geary_autocorrelation:
        testing correct protpy Geary Autocorrelation functionality.
    """
    def setUp(self):
        """ Import protein sequences from test fasta files using Biopython package. """
        #using next() to get first item (protein seq) from SeqIO Generator
        with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
            self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
            self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
        with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
            self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
            self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

        #list of canonical amino acids
        self.amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
            "Q", "R", "S", "T", "V", "W", "Y"]

    def test_moreaubroto_autocorrelation(self):
        """ Testing moreaubroto autocorrelation descriptor attributes and functionality. """
#1.)
        moreaubroto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moreaubroto_lag:
            moreaubroto_seq1 = protpy.moreaubroto_autocorrelation(self.protein_seq1, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moreaubroto_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moreaubroto_seq1.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moreaubroto_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moreaubroto_seq1.columns):
                self.assertTrue(bool(re.match(r"MoreauBrotoAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))     
#2.)
        moreaubroto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moreaubroto_lag:
            moreaubroto_seq2 = protpy.moreaubroto_autocorrelation(self.protein_seq2, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moreaubroto_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moreaubroto_seq2.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moreaubroto_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moreaubroto_seq2.columns):
                self.assertTrue(bool(re.match(r"MoreauBrotoAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#3.)
        moreaubroto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moreaubroto_lag:
            moreaubroto_seq3 = protpy.moreaubroto_autocorrelation(self.protein_seq3, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moreaubroto_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moreaubroto_seq3.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moreaubroto_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moreaubroto_seq3.columns):
                self.assertTrue(bool(re.match(r"MoreauBrotoAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#4.)
        moreaubroto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moreaubroto_lag:
            moreaubroto_seq4 = protpy.moreaubroto_autocorrelation(self.protein_seq4, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moreaubroto_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moreaubroto_seq4.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moreaubroto_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moreaubroto_seq4.columns):
                self.assertTrue(bool(re.match(r"MoreauBrotoAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            moreaubroto_seq5 = protpy.moreaubroto_autocorrelation(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            moreaubroto_seq6 = protpy.moreaubroto_autocorrelation(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            moreaubroto_seq7 = protpy.moreaubroto_autocorrelation(invalid_seq7)


    def test_moran_autocorrelation(self):
        """ Testing moran autocorrelation descriptor attributes and functionality. """
#1.)
        moran_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moran_auto_lag:
            moran_auto_seq1 = protpy.moran_autocorrelation(self.protein_seq1, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moran_auto_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moran_auto_seq1.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moran_auto_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moran_auto_seq1.columns):
                self.assertTrue(bool(re.match(r"MoranAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#2.)
        moran_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moran_auto_lag:
            moran_auto_seq2 = protpy.moran_autocorrelation(self.protein_seq2, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moran_auto_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moran_auto_seq2.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moran_auto_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moran_auto_seq2.columns):
                self.assertTrue(bool(re.match(r"MoranAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#3.)
        moran_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moran_auto_lag:
            moran_auto_seq3 = protpy.moran_autocorrelation(self.protein_seq3, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moran_auto_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moran_auto_seq3.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moran_auto_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moran_auto_seq3.columns):
                self.assertTrue(bool(re.match(r"MoranAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}".format(col))
#4.)
        moran_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in moran_auto_lag:
            moran_auto_seq4 = protpy.moran_autocorrelation(self.protein_seq4, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(moran_auto_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(moran_auto_seq4.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(moran_auto_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(moran_auto_seq4.columns):
                self.assertTrue(bool(re.match(r"MoranAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}".format(col))
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            moran_auto_seq5 = protpy.moran_autocorrelation(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            moran_auto_seq6 = protpy.moran_autocorrelation(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            moran_auto_seq7 = protpy.moran_autocorrelation(invalid_seq7)

    def test_geary_autocorrelation(self):
        """ Testing geary autocorrelation descriptor attributes and functionality. """
#1.)
        geary_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in geary_auto_lag:
            geary_auto_seq1 = protpy.geary_autocorrelation(self.protein_seq1, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(geary_auto_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(geary_auto_seq1.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(geary_auto_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(geary_auto_seq1.columns):
                self.assertTrue(bool(re.match(r"GearyAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#2.)
        geary_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in geary_auto_lag:
            geary_auto_seq2 = protpy.geary_autocorrelation(self.protein_seq2, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(geary_auto_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(geary_auto_seq2.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(geary_auto_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(geary_auto_seq2.columns):
                self.assertTrue(bool(re.match(r"GearyAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))
#3.)
        geary_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in geary_auto_lag:
            geary_auto_seq3 = protpy.geary_autocorrelation(self.protein_seq3, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(geary_auto_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(geary_auto_seq3.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(geary_auto_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(geary_auto_seq3.columns):
                self.assertTrue(bool(re.match(r"GearyAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}".format(col))
#4.)
        geary_auto_lag = list(range(5, 35, 5))
        properties = ["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102", 
            "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]

        for lag in geary_auto_lag:
            geary_auto_seq4 = protpy.geary_autocorrelation(self.protein_seq4, lag=lag, 
                properties=properties, normalize=True)
            self.assertIsInstance(geary_auto_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(geary_auto_seq4.shape, (1, lag*len(properties)), 'Descriptor not of correct shape.') 
            self.assertTrue(geary_auto_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')        

            #check all columns follow pattern of MoreauBrotoAuto_X_Y where x is the asscession number of
            #   the AAindex record and y is the count of the descriptor
            for col in list(geary_auto_seq4.columns):
                self.assertTrue(bool(re.match(r"GearyAuto_[A-Z0-9]{10}_[0-9]", col)), 
                    "Column name doesn't match expected regex pattern: {}".format(col))
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            geary_auto_seq5 = protpy.geary_autocorrelation(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            geary_auto_seq6 = protpy.geary_autocorrelation(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            geary_auto_seq7 = protpy.geary_autocorrelation(invalid_seq7)


###########################################################################
#################            protPy Module Tests          #################
###########################################################################

import pandas as pd
import numpy as np
import os
import unittest
import re
from Bio import SeqIO
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import protpy as protpy

class ProtPyTests(unittest.TestCase):
    """
    Test suite for testing protpy package. 

    Test Cases
    ----------
    test_protpy_metadata:
        testing pypi software package metadata.
    test_valid_descriptors:
        testing correct list of valid descriptors.
    """
    def setUp(self):
        """ Import protein sequences from test fasta files using Biopython package. """
        #using next() to get first item (protein seq) from SeqIO Generator
        with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
            self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
            self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
        with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
            self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
            self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

        #list of canonical amino acids
        self.amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
            "Q", "R", "S", "T", "V", "W", "Y"]

    def test_protpy_metadata(self):
        """ Testing correct protpy version and metadata. """
        self.assertEqual(protpy.__version__, "0.0.1", 
            "protpy version is not correct, got: {}".format(protpy.__version__))
        self.assertEqual(protpy.__name__, "protpy", 
            "protpy software name is not correct, got: {}".format(protpy.__name__))
        self.assertEqual(protpy.__author__, "AJ McKenna, https://github.com/amckenna41", 
            "protpy author is not correct, got: {}".format(protpy.__author__))
        self.assertEqual(protpy.__authorEmail__, "amckenna41@qub.ac.uk", 
            "protpy author email is not correct, got: {}".format(protpy.__authorEmail__))
        self.assertEqual(protpy.__url__, "https://github.com/amckenna41/protPy", 
            "protpy repo URL is not correct, got: {}".format(protpy.__url__))
        self.assertEqual(protpy.__download_url__, "https://github.com/amckenna41/protPy/archive/refs/heads/main.zip", 
            "protpy repo download URL is not correct, got: {}".format(protpy.__download_url__))
        self.assertEqual(protpy.__status__, "Production", 
            "protpy status is not correct, got: {}".format(protpy.__status__))
        self.assertEqual(protpy.__license__, "MIT", 
            "protpy license type is not correct, got: {}".format(protpy.__license__))
        self.assertEqual(protpy.__maintainer__, "AJ McKenna", 
            "protpy maintainer is not correct, got: {}".format(protpy.__license__))
        self.assertEqual(protpy.__keywords__, ["bioinformatics", "protein engineering", 
            "python", "pypi", "machine learning", "aaindex", "protein descriptors", 
            "physiochemical descriptors", "biochemical descriptors" "structural descriptors"], 
            "protpy keywords is not correct, got: {}".format(protpy.__keywords__))

    def test_valid_descriptors(self):
        """ Testing correct list of available descriptors is included in package attribute. """
        self.assertEqual(len(protpy.all_descriptors), 14, 
            "Expected there to be 14 total descriptors, got {}".format(len(protpy.all_descriptors)))
        self.assertIsInstance(protpy.all_descriptors, list, 
            "all_descriptors should be of type list, got {}.".format(type(protpy.all_descriptors)))
        self.assertIn('aa_composition', protpy.all_descriptors, 
            "aa_composition should be in available descriptors list.")
        self.assertIn('dipeptide_composition', protpy.all_descriptors, 
            "dipeptide_composition should be in available descriptors list.")
        self.assertIn('tripeptide_composition', protpy.all_descriptors,
            "tripeptide_composition should be in available descriptors list.")
        self.assertIn('pseudo_amino_acid_composition', protpy.all_descriptors,
            "pseudo_amino_acid_composition should be in available descriptors list.")
        self.assertIn('amphiphilic_pseudo_amino_acid_composition', protpy.all_descriptors, 
            "amphiphilic_pseudo_amino_acid_composition should be in available descriptors list.")
        self.assertIn('moreaubroto_autocorrelation', protpy.all_descriptors,
            "moreaubroto_autocorrelation should be in available descriptors list.")
        self.assertIn('moran_autocorrelation', protpy.all_descriptors, 
            "moran_autocorrelation should be in available descriptors list.")
        self.assertIn('geary_autocorrelation', protpy.all_descriptors,
            "geary_autocorrelation should be in available descriptors list.")
        self.assertIn('conjoint_triad', protpy.all_descriptors,
            "conjoint_triad should be in available descriptors list.")
        self.assertIn('ctd_composition', protpy.all_descriptors,
            "ctd_composition should be in available descriptors list.")
        self.assertIn('ctd_transition', protpy.all_descriptors,
            "ctd_transition should be in available descriptors list.")
        self.assertIn('ctd_distribution', protpy.all_descriptors,
            "ctd_distribution should be in available descriptors list.")
        self.assertIn('sequence_order_coupling_number', protpy.all_descriptors,
            "sequence_order_coupling_number should be in available descriptors list.")
        self.assertIn('quasi_sequence_order', protpy.all_descriptors,
            "quasi_sequence_order should be in available descriptors list.")



################################################################################
#############                   Composition                       ##############
################################################################################

import re
from typing import Type
import pandas as pd
import math
from aaindex import aaindex1

#list of amino acids
amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
    "Q", "R", "S", "T", "V", "W", "Y"]

def aa_composition(sequence):
    """
    Calculate Amino Acid Composition (AAComp) of protein sequence. AAComp
    describes the fraction of each amino acid type within a protein sequence,
    and is calculated as:

    AA_Comp(s) = AA(t)/N(s)

    where AA_Comp(s) is the AAComp of protein sequence s, AA(t) is the number
    of amino acid types t (where t = 1,2,..,20) and N(s) is the length of the
    sequence s.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.

    Returns
    -------
    :amino_acid_composition_df : pd.DataFrame
        pandas dataframe of AAComp for protein sequence. Dataframe will
        be of the shape 20 x 1, where 20 is the number of features 
        calculated from the descriptor (for the 20 amino acids).
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    amino_acid_composition = {}
    for aa in amino_acids:
        amino_acid_composition[aa] = round(float(sequence.count(aa)) / len(sequence) * 100, 3)

    #transform values and columns to DataFrame
    amino_acid_composition_df = pd.DataFrame([list(amino_acid_composition.values())], columns=list(amino_acid_composition.keys()))

    return amino_acid_composition_df

def dipeptide_composition(sequence):
    """
    Calculate Dipeptide Composition (DPComp) for protein sequence.
    Dipeptide composition is the fraction of each dipeptide type within a
    protein sequence. With dipeptides being of length 2 and there being 20
    canonical amino acids this creates 20^2 different combinations, thus a
    400-Dimensional vector will be produced such that:

    DPComp(s,t) = AA(s,t) / N -1

    where DPComp(s,t) is the dipeptide composition of the protein sequence
    for amino acid type s and t (where s and t = 1,2,..,20), AA(s,t) is the number
    of dipeptides represented by amino acid type s and t and N is the total number
    of dipeptides.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.

    Returns
    -------
    :dipeptide_composition_df : pd.DataFrame
        pandas dataframe of dipeptide composition for protein sequence. Dataframe will
        be of the shape 400 x 1, where 400 is the number of features calculated 
        from the descriptor (20^2 for the 20 canonical amino acids).
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    dipeptide_composition = {}
    for i in amino_acids:
        for j in amino_acids:
            dipep = i + j
            dipeptide_composition[dipep] = round(
                float(sequence.count(dipep)) / (len(sequence)-1) * 100, 2
            )

    #transform values and columns to DataFrame
    dipeptide_composition_df = pd.DataFrame([list(dipeptide_composition.values())], columns=list(dipeptide_composition.keys()))

    return dipeptide_composition_df

def tripeptide_composition(sequence):
    """
    Calculate Tripeptide Composition (TPComp) of protein sequence.
    Tripeptide composition is the fraction of each tripeptide type within a
    protein sequence. With tripeptides being of length 3 and there being 20
    canonical amino acids this creates 20^3 different combinations, thus a
    8000-Dimensional vector will be produced such that:

    TPComp(s,t,u) = AA(s,t,u) / N -1

    where TPComp(s,t,u) is the tripeptide composition of the protein sequence
    for amino acid type s, t and u (where s, t and u = 1,2,..,20), AA(s,t,u) is
    the number of tripeptides represented by amino acid type s and t, and N is
    the total number of tripeptides.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.

    Returns
    -------
    :tripeptide_composition_df : pd.DataFrame
        pandas DataFrame of tripeptide composition for protein sequence. Dataframe will
        be of the shape 8000 x 1, where 8000 is the number of features calculated 
        from the descriptor (20^3 for the 20 canonical amino acids).
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    tripeptide_composition = {}
    tripeptides = []    

    #get list of tripeptides
    for i in amino_acids:
        for j in amino_acids:
            for k in amino_acids:
                tripeptides.append(i + j + k)

    #get frequency of each tripeptide in the sequence
    for i in tripeptides:
        tripeptide_composition[i] = len(re.findall(i, sequence))

    #transform values and columns to DataFrame
    tripeptide_composition_df = pd.DataFrame([list(tripeptide_composition.values())], columns=list(tripeptide_composition.keys()))

    return tripeptide_composition_df

def pseudo_amino_acid_composition(sequence, lamda=30, weight=0.05, properties=["ARGP820101", "KUHL950101"]):
    """
    Calculate Pseudo Amino Acid Composition features for the protein sequence.
    Similar to the quasi-sequence order descriptor, the pseudo amino acid descriptor is
    made up of a 50-dimensional vector in which the first 20 components are a weighted
    sum of the amino acid composition and 30 are physiochemical square correlations as
    dictated by the lamda and properties parameters. This generates an output of 
    [(20 + lamda), 1] = 50 x 1 when using the default lamda of 30. By default, the 
    physiochemical properties used are  hydrophobicity (ARGP820101) and hydrophillicity 
    (KUHL950101) indices, with a lamda of 30 and weight of 0.05.

    Parameters 
    ----------
    :sequence : str
        protein sequence in str form.
    :lamda : int (default = 30)
        lamda parameter that reflects the rank correlation and should be a non-negative
        integer and not larger than the length of the protein sequence.
    :weight : float (default = 0.05)
        weighting factor allowing for weights to be added to the additioanl descriptor
        components with respect to the conventional amino acid components. 
    :properties : str/array (default=["ARGP820101", "KUHL950101"])
        single or multiple amino acid index properties from the AAI database used for 
        calculating the sequence-order.

    Returns
    -------
    :pseudo_amino_acid_composition_df : pd.Dataframe
        pandas Dataframe of pseudo amino acid composition for protein sequence. Dataframe will
        be of the shape [(20 + lamda),1] - 50 x 1, where 50 is the number of features calculated 
        from the descriptor, when using the default lamda of 30.

    References
    ----------
    [1]: Chou, K. C. (2001). Prediction of protein cellular attributes using pseudo-amino acid 
         composition. Proteins, 43(3), 246–255. https://doi.org/10.1002/prot.1035
    [2]: Chen C, Zhou X, Tian Y, Zhou X, Cai P: Predicting protein structural class with pseudo-amino 
         acid composition and support vector machine fusion network. Anal Biochem 2006, 
         357: 116–121. 10.1016/j.ab.2006.07.022
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    #set lamda to its default value if <0, or > sequence len or not an int
    if ((lamda < 0) or (lamda > len(sequence)) or not isinstance(lamda, int)):
        lamda = 30

    #ensure at least 1 property input to function and or properties is a list so it can be iterated over
    if (properties == "" or properties == []):
        raise ValueError('At least one property value must be input to function.')
    if (isinstance(properties, str)):   #cast properties to list if str
        properties = [properties]

    #initialise dicts to store AAI properties and values
    aai_properties = {}
    for prop in properties:
        aai_properties[prop] = {}

    #iterate through list of properties, getting property values from AAIndex
    for prop in properties:

        #get property values from AAIndex & reshape
        aaindex1[prop]['values'].pop('-', None)
        prop_aminoacid_values = aaindex1[prop]['values']
     
        #normalise property values, calculate mean and std dev
        aai_property_vals = {}
        for i, j in prop_aminoacid_values.items():
            aai_property_vals[i] = (j - (sum(prop_aminoacid_values.values()) / len(prop_aminoacid_values.values()))) / _std(prop_aminoacid_values.values(), ddof=0)

        aa_counter = 0
        #assign property and associated amino acid values to aai_property_vals array
        for i, j in aai_property_vals.items():
            aai_property_vals[amino_acids[aa_counter]] = aai_property_vals[i]
            aa_counter+=1
        aai_properties[prop] = aai_property_vals

    #### Pseudo AAC 1 ####

    #calculate pseudo AAC for sequence
    right_part = 0.0
    for i in range(lamda):
        right_part = right_part + sequence_order_correlation_factor(
            sequence, i + 1, aai_properties
        )

    #get amino acid composition
    aa_comp = aa_composition(sequence)    
    result = {}
    #applying weighting factor to components
    temp = 1 + weight * right_part

    #append each descriptor feature value to results dict
    for index, i in enumerate(amino_acids):
        result["PseudoAAC1_" + str(index + 1)] = round(aa_comp.iloc[0][i] / temp, 3) 

    ##### Pseudo AAC 2 ####

    #calculate pseudo AAC for sequence 
    right_part = []
    for i in range(lamda):
        right_part.append(sequence_order_correlation_factor(sequence, i + 1, aai_properties))

    #applying weighting factor to components
    temp = 1 + weight * sum(right_part)

    #append each descriptor feature value to results dict
    for index in range(20, 20 + lamda):
        result["PseudoAAC2_" + str(index + 1)] = round(
            weight * right_part[index - 20] / temp * 100, 3
        )

    #convert resultant descriptor values into a dataframe
    pseudo_amino_acid_composition_df = pd.DataFrame([list(result.values())], columns=list(result.keys()))

    return pseudo_amino_acid_composition_df

def sequence_order_correlation_factor(sequence, k=1, properties=[]):
    """
    Calculate the sequence order correlation factor with gap = k for the inputted
    physiochemical properties. 

    Parameters
    ----------
    :k : int (default = 1)
        gap in sequence for calculating factor.
    :properties : list (default = [])
        list of physiochemical properties.

    Returns
    -------
    :seqOrderCorrelationFactor : float
        sequence order correlation factor with gap = k.
    
    References
    ----------
    [1]: Manish C. Saraf, Gregory L. Moore, Costas D. Maranas, Using multiple 
         sequence correlation analysis to characterize functionally important 
         protein regions, Protein Engineering, Design and Selection, Volume 16, 
         Issue 6, June 2003, Pages 397–406, https://doi.org/10.1093/protein/gzg053
    
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    #ensure at least 1 property input to function and or properties is a list so it can be iterated over
    if (properties == "" or properties == []):
        raise ValueError('At least one property value must be input to function.')

    #cast properties to list if str
    if (isinstance(properties, str)):   
        properties = [properties]

    #iterate through sequence, calculate sequence order correlation for amino acids    
    res = []
    for i in range(len(sequence) - k):
        AA1 = sequence[i]
        AA2 = sequence[i + k]
        res.append(correlation_function(AA1, AA2, properties))
    
    #get resultant correlation values
    result = round(sum(res) / (len(sequence) - k), 3)

    return result

def correlation_function(aa1, aa2, properties):
    """
    Calculate the correlation between two amino acids using the selected
    properties for APAAC (type II PseAAC).

    Parameters
    ----------
    :aa1 : str 
        amino aicd letter.
    :aa2: str 
        amino acid letter.
    :properties : list
        list of physiochemical properties.

    Returns
    -------
    :seqOrderCorrelationFactor : float
        sequence order correlation factor with gap = k.    
    """
    #if only 1 property passed in, cast to a list
    if isinstance(properties, str):
        properties = [properties]

    theta = 0.0
    #calculate correlation between 2 input amino acids per property
    for prop in properties:
        temp = normalize_aap(properties[prop])
        theta = theta + math.pow(temp[aa1] - temp[aa2], 2)
    
    result = round(theta / len(properties), 3)
    return result

def normalize_aap(property):
    """
    Normalize and centralize the amino acid property values.

    Parameters
    ----------
    :property : dict 
        dictionary of amino acid property and associated values.

    Returns
    -------
    :result : dict
        normalized amino acid property values.    
    """
    #normalize amino acid property values
    result = {}
    for i, j in property.items():
        result[i] = (j - (sum(property.values()) / len(property.values())) / _std(property.values(), ddof=0))

    return result

def amphiphilic_pseudo_aac(sequence, lamda=30, weight=0.5, properties=["ARGP820101", "KUHL950101"]):
    """
    Calculate the Type 2 Pseudo AA Composition descriptors from the input physiochemical
    properties. The number of features generated depends on the property and lambda value: 
    20 + i*lambda discrete numbers are generated to represent a protein, where i is the number of 
    amino acid attributes selected). By default, the ARGP820101 (Hydrophobicity) and KUHL950101
    (Hydrophilicity) properties are used from the AAIndex. More about the descriptor can be found
    here [3]. 

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.
    :lamda : int (default = 30)
        lamda parameter that reflects the rank correlation and should be a non-negative
        integer and not larger than the length of the protein sequence.
    :weight : float (default = 0.05)
        weighting factor allowing for weights to be added to the additioanl descriptor
        components with respect to the conventional amino acid components. 
    :properties : str/array (default =["ARGP820101", "KUHL950101"])
        single or multiple amino acid index properties from the AAI database.

    Returns
    -------
    :amp_pseudo_amino_acid_composition_df : pd.Dataframe
        pandas Dataframe of amphiphilic pseudo amino acid composition for protein sequence. Dataframe will
        be of the shape (20 +(i*lambda), 1), where (20 + (i*lambda)) is the number of features calculated 
        from the descriptor. 
    
    References
    ----------
    [1]: Kuo-Chen Chou. Using amphiphilic pseudo amino acid composition to predict enzyme
         subfamily classes. Bioinformatics, 2005,21,10-19.
    [2]: Chou,K.C. and Cai Y.D. (2005). Prediction of membrane protein types by incorporating 
         amphipathic effects, J Chem Inf Model, 45(2):407-13
    [3]: http://www.csbio.sjtu.edu.cn/bioinf/PseAAC/type2.htm
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))
            
    #set lamda to its default value if <0, or > sequence len or not an int
    if ((lamda < 0) or (lamda > len(sequence)) or not isinstance(lamda, int)):
        lamda = 30

    #keys of dicts should be AA not properties
    aai_properties = {}
    aai_property_vals = {}

    #ensure properties is a list so it can be iterated over
    if (isinstance(properties, list) or len(properties) == 1):
        properties = [properties]

    #get amino acid values from AAI for property 
    for prop in properties: 
        aai_properties[prop] = aaindex.get_values_from_record(prop)
    right_part = 0.0
    for i in range(lamda):
        right_part = right_part + sum(
            amphiphilic_sequence_order_correlation_factor(sequence, k=i + 1)
        )
    aa_comp = aa_composition(sequence)

    amp_pseudo_amino_acid_composition = {}
    temp = 1 + weight * right_part
    for index, i in enumerate(amino_acids):
        amp_pseudo_amino_acid_composition["APAAC" + str(index + 1)] = round(aa_comp[i] / temp, 3)

    #transform descriptor data into pandas dataframe
    amp_pseudo_amino_acid_composition_df = pd.DataFrame([list(amp_pseudo_amino_acid_composition.values())], columns=list(amp_pseudo_amino_acid_composition.keys()))

    return amp_pseudo_amino_acid_composition_df

def _std(array, ddof=1):
    """
    Calculate the standard deviation of the array data.

    Parameters
    ----------
    :array : np.array
        numpy array of floats.

    Returns
    -------
    :res : np.array
        input array after standard deviation transformation.
    """
    return math.sqrt(sum([math.pow(i - sum(array) / len(array), 2) for i in array]) 
        / (len(array) - ddof))


###########################################################################
#############      protPy - Quasi Sequence Module Tests      ##############
###########################################################################

import pandas as pd
import numpy as np
import os
import unittest
import re
from decimal import *
from Bio import SeqIO
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import protpy as protpy

class ProtPyQuasiSequenceOrderTests(unittest.TestCase):
    """
    Test suite for testing CTD (Composition, Transition, Distribution) 
    module and functionality in protpy package. 

    Test Cases
    ----------
    test_sequence_order_coupling_number:
        testing correct protpy sequence order coupling number functionality.
    test_quasi_sequence_order:
        testing correct protpy quasi sequence order functionality.
    """
    def setUp(self):
        """ Import protein sequences from test fasta files using Biopython package. """
        #using next() to get first item (protein seq) from SeqIO Generator
        with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
            self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
            self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
        with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
            self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
            self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

    def test_sequence_order_coupling_number(self):
        """ Testing sequence order coupling number descriptor attributes and functionality. """
        socn_lag = list(range(5, 35, 5))
        distance_matrix_schneider_wrede = "schneider-wrede-physiochemical-distance-matrix"
        distance_matrix_grantham = "grantham-physiochemical-distance-matrix"
#1.)
        for lag in socn_lag:
            socn_seq1 = protpy.seq_order_coupling_number(self.protein_seq1, lag=lag, 
                distance_matrix=distance_matrix_schneider_wrede)

            self.assertFalse(socn_seq1.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(socn_seq1.shape, (1, lag), 'Descriptor not of correct shape.') 
            self.assertIsInstance(socn_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
            for col in list(socn_seq1.columns):
                #check all columns follow pattern of SOCNX or SOCNXY where x & y integers between 0 and 9
                self.assertTrue((bool(re.match(r'SOCN[0-9]{1}$', col)) or bool(re.match(r'SOCN[0-9]{2}$', col))), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))     
            self.assertTrue(socn_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
            self.assertTrue(all(col == np.float64 for col in list(socn_seq1.dtypes)), "")
#1.2)
            socn_seq1 = protpy.seq_order_coupling_number(self.protein_seq1, lag=lag, 
                distance_matrix=distance_matrix_grantham)

            self.assertFalse(socn_seq1.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(socn_seq1.shape, (1, lag), 'Descriptor not of correct shape.') 
            self.assertIsInstance(socn_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
            for col in list(socn_seq1.columns):
                #check all columns follow pattern of SOCNX or SOCNXY where x & y integers between 0 and 9
                self.assertTrue((bool(re.match(r'SOCN[0-9]{1}$', col)) or bool(re.match(r'SOCN[0-9]{2}$', col))), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))     
            self.assertTrue(socn_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
            self.assertTrue(all(col == np.float64 for col in list(socn_seq1.dtypes)), "")

#2.)
            socn_seq2 = protpy.seq_order_coupling_number(self.protein_seq2, lag=lag, 
                distance_matrix=distance_matrix_schneider_wrede)

            self.assertFalse(socn_seq2.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(socn_seq2.shape, (1, lag), 'Descriptor not of correct shape.') 
            self.assertIsInstance(socn_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
            for col in list(socn_seq2.columns):
                #check all columns follow pattern of SOCNX or SOCNXY where x & y integers between 0 and 9
                self.assertTrue((bool(re.match(r'SOCN[0-9]{1}$', col)) or bool(re.match(r'SOCN[0-9]{2}$', col))), 
                    "Column name doesn't match expected regex pattern: {}.".format(col))     
            self.assertTrue(socn_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
            self.assertTrue(all(col == np.float64 for col in list(socn_seq2.dtypes)), "")
#3.)

#4.)

#5.)
    def test_quasi_sequence_order(self):
        """ Testing quasi sequence order descriptor attributes and functionality. """
#1.)

#2.)

#3.)

#4.)

#5.)
        pass


# def seq_order_coupling_number(sequence, lag=30,
#     distance_matrix="schneider-wrede-physiochemical-distance-matrix.json"):

# def quasi_sequence_order(sequence, max_lag=30, weight=0.1,
#     distance_matrix="schneider-wrede-physicochemical-distance-matrix.json"):


################################################################################
############                    Conjoint Triad                     #############
################################################################################

import pandas as pd

#list of amino acids
amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
    "Q", "R", "S", "T", "V", "W", "Y"]
    
#amino acid triads
aa_triads = {
    1: ["A", "G", "V"],
    2: ["I", "L", "F", "P"],
    3: ["Y", "M", "T", "S"],
    4: ["H", "N", "Q", "W"],
    5: ["R", "K"],
    6: ["D", "E"],
    7: ["C"],
}

def conjoint_triad(sequence):
    """
    Calculate Conjoint Triad features (CTriad) for a protein sequence. CTF
    mainly considers neighbor relationships in protein sequences by encoding
    each protein sequence using the triad (continuous three amino acids)
    frequency distribution extracted from a 7-letter reduced alphabet [7]. This
    descriptor calculates 343 different features (7x7x7), with the output
    being of shape 343 x 1 for a sequence.

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.

    Returns
    -------
    :conj_triad_df : pd.Dataframe
        pandas Dataframe of CTriad descriptor values for all protein sequences. Dataframe
        will be of the shape 343 x 1, where 343 is the number of features calculated 
        from the descriptor for a sequence.

    References
    ----------
    [1]: J. Shen et al., “Predicting protein-protein interactions based only on sequences
        information,” Proc. Natl. Acad. Sci. U. S. A., vol. 104, no. 11, pp. 4337–4341, 2007.
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))
    
    #if invalid amino acids in sequence, raise value error
    for aa in sequence:
        if (aa not in amino_acids):
            raise ValueError("Invalid amino acid in protein sequence: ".format(aa))

    conj_triad = {}
    _aa_triads = {}

    #expand aa_triads into form {AminoAcid: triad}
    for i in aa_triads:
        for j in aa_triads[i]:
            _aa_triads[j] = i

    #get protein number for each triad
    protein_num = sequence
    for i in _aa_triads:
        protein_num = protein_num.replace(i, str(_aa_triads[i]))

    #calculate 7 triads for amino acids
    for i in range(1, 8):
        for j in range(1, 8):
            for k in range(1, 8):
                temp = str(i) + str(j) + str(k)
                conj_triad[temp] = protein_num.count(temp)

    #transform descriptor values to dataframe
    conj_triad_df = pd.DataFrame([list(conj_triad.values())], columns=list(conj_triad.keys()))

    return conj_triad_df



###########################################################################
#############      protPy - Conjoint Triad Module Tests      ##############
###########################################################################

import pandas as pd
import numpy as np
import os
import unittest
import re
from Bio import SeqIO
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import protpy as protpy

class ProtPyConjointTriadTests(unittest.TestCase):
    """
    Test suite for testing conjoint triad module and functionality 
    in protpy package. 

    Test Cases
    ----------
    test_conjoint_triad_composition:
        testing correct protpy conjoint triad functionality.
    """
    def setUp(self):
        """ Import protein sequences from test fasta files using Biopython package. """
        #using next() to get first item (protein seq) from SeqIO Generator
        with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
            self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
            self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
        with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
            self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
            self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

    def test_conjoint_triad(self):
        """ Testing conjoing triad descriptor attributes and functionality. """
#1.)
        conjoint_triad_seq1 = protpy.conjoint_triad(self.protein_seq1)

        print("conjoint_triad_seq1")
        print(conjoint_triad_seq1.columns)
        self.assertFalse(conjoint_triad_seq1.empty, 'Descriptor dataframe should not be empty')
        self.assertEqual(conjoint_triad_seq1.shape, (1, 343), 'Descriptor not of correct shape.') 
        self.assertIsInstance(conjoint_triad_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertTrue(conjoint_triad_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')    

        for col in list(conjoint_triad_seq1.columns):
            self.assertTrue(bool(re.match(r"[0-9]{3}", col)), 
                "Column name doesn't match expected regex pattern: {}.".format(col))   
#2.)
        conjoint_triad_seq2 = protpy.conjoint_triad(self.protein_seq2)

        self.assertFalse(conjoint_triad_seq2.empty, 'Descriptor dataframe should not be empty')
        self.assertEqual(conjoint_triad_seq2.shape, (1, 343), 'Descriptor not of correct shape.') 
        self.assertIsInstance(conjoint_triad_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertTrue(conjoint_triad_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.') 

        for col in list(conjoint_triad_seq2.columns):
            self.assertTrue(bool(re.match(r"[0-9]{3}", col)), 
                "Column name doesn't match expected regex pattern: {}.".format(col))   
#3.)
        conjoint_triad_seq3 = protpy.conjoint_triad(self.protein_seq3)

        self.assertFalse(conjoint_triad_seq3.empty, 'Descriptor dataframe should not be empty')
        self.assertEqual(conjoint_triad_seq3.shape, (1, 343), 'Descriptor not of correct shape.') 
        self.assertIsInstance(conjoint_triad_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertTrue(conjoint_triad_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.') 

        for col in list(conjoint_triad_seq3.columns):
            self.assertTrue(bool(re.match(r"[0-9]{3}", col)), 
                "Column name doesn't match expected regex pattern: {}.".format(col))   
#4.)
        conjoint_triad_seq4 = protpy.conjoint_triad(self.protein_seq4)

        self.assertFalse(conjoint_triad_seq4.empty, 'Descriptor dataframe should not be empty')
        self.assertEqual(conjoint_triad_seq4.shape, (1, 343), 'Descriptor not of correct shape.') 
        self.assertIsInstance(conjoint_triad_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertTrue(conjoint_triad_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.') 

        for col in list(conjoint_triad_seq4.columns):
            self.assertTrue(bool(re.match(r"[0-9]{3}", col)), 
                "Column name doesn't match expected regex pattern: {}.".format(col))   
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            conjoint_triad_seq5 = protpy.conjoint_triad(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            conjoint_triad_seq6 = protpy.conjoint_triad(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            conjoint_triad_seq7 = protpy.conjoint_triad(invalid_seq7)


# ###########################################################################
# ##################      protPy - CTD Module Tests      ####################
# ###########################################################################

# import pandas as pd
# import numpy as np
# import os
# import unittest
# import re
# from Bio import SeqIO
# unittest.TestLoader.sortTestMethodsUsing = None
# from json import JSONDecodeError

# import protpy as protpy

# class ProtPyCTDTests(unittest.TestCase):
#     """
#     Test suite for testing CTD (Composition, Transition, Distribution) 
#     module and functionality in protpy package. 

#     Test Cases
#     ----------
#     test_ctd:
#         testing correct protpy CTD functionality.
#     test_ctd_composition:
#         testing correct protpy composition functionality.
#     test_ctd_transition:
#         testing correct protpy transition functionality.
#     test_ctd_distribution:
#         testing correct protpy distribution functionality.
#     """
#     def setUp(self):
#         """ Import protein sequences from test fasta files using Biopython package. """
#         #using next() to get first item (protein seq) from SeqIO Generator
#         with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
#             self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

#         with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
#             self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
#         with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
#             self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

#         with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
#             self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

#     def test_ctd(self):
#         """ Testing CTD descriptor attributes and functionality. """   
#         properties = ["hydrophobicity", "normalized_VDWV", "polarity", "charge",
#             "sec_struct", "solvent_accessibility", "polarizability"]
 
#         for prop in properties:
# #1.)
#             ctd_seq1 = protpy.ctd_(self.protein_seq1, property=prop)
            
#             # print("ctd_seq1")
#             # print(ctd_seq1)
            
#         # hydrophobicity_CTD_C_1
#         #                 self.assertTrue(bool(re.match(r"MoreauBrotoAuto_[A-Z0-9]{10}_[0-9]", col)), 

#             self.assertFalse(ctd_seq1.empty, 'Descriptor dataframe should not be empty')
#             self.assertEqual(ctd_seq1.shape, (1, 147), 'Descriptor not of correct shape.') 
#             self.assertIsInstance(ctd_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
#             self.assertTrue(ctd_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  

# # 'hydrophobicity_CTD_D_2_001', 

#             print(ctd_seq1.columns)
#             print(prop)
#             for col in list(ctd_seq1.columns):
                
#                 # print("col", col)
#                 # print(bool(re.search(prop + "_CTD_[A-Z]{1}_[0-9]", col)))

#                 # print("prop", prop)
#                 self.assertTrue((bool(re.search(prop + "_CTD_[A-Z]{1}_[0-9]", col)) or bool(re.search(prop + "_CTD_[A-Z]{1}_[0-9]_[0-9]{3}", col))), 
#                     "Column name doesn't match expected regex pattern: {}, {}.".format(col, prop))  

# #                 # r'\b(?<=\w){0}\b(?!\w)' re.search('(.+)'+var_name+'(.+)'
# #                 # self.assertTrue(bool(re.match(rf"{re.escape(prop)}_CTD_[A-Z]{1}_[0-9]", col)), 
# #                 #     "Column name doesn't match expected regex pattern: {}.".format(col))  

# # #2.)
# #             ctd_composition_seq1 = protpy.ctd_composition(self.protein_seq1, property=prop)


# #             self.assertFalse(ctd_composition_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_composition_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_composition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_composition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  

# # #3.)
# #             ctd_transition_seq1 = protpy.ctd_transition(self.protein_seq1, property=prop)


# #             self.assertFalse(ctd_transition_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_transition_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_transition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_transition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  

# # #4.)
# #             ctd_distribution_seq1 = protpy.ctd_distribution(self.protein_seq1, property=prop)


#             # self.assertFalse(ctd_distribution_seq1.empty, 'Descriptor dataframe should not be empty')
#             # self.assertEqual(ctd_distribution_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
#             # self.assertIsInstance(ctd_distribution_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
#             # self.assertTrue(ctd_distribution_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  








# # #2.)        
# #         for prop in properties:
# #             ctd_seq1 = protpy.ctd_(self.protein_seq1, property=prop)

# #             self.assertFalse(ctd_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  


# #             ctd_composition_seq1 = protpy.ctd_composition(self.protein_seq1, property=prop)


# #             self.assertFalse(ctd_composition_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_composition_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_composition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_composition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  


# #             ctd_transition_seq1 = protpy.ctd_transition(self.protein_seq1, property=prop)


# #             self.assertFalse(ctd_transition_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_transition_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_transition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_transition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  


# #             ctd_distribution_seq1 = protpy.ctd_distribution(self.protein_seq1, property=prop)


# #             self.assertFalse(ctd_distribution_seq1.empty, 'Descriptor dataframe should not be empty')
# #             self.assertEqual(ctd_distribution_seq1.shape, (1, 3), 'Descriptor not of correct shape.') 
# #             self.assertIsInstance(ctd_distribution_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
# #             self.assertTrue(ctd_distribution_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')  


# # def seq_order_coupling_number(sequence, lag=30,
# #     distance_matrix="schneider-wrede-physiochemical-distance-matrix.json"):

# # hydrophobicity = {"name": "hydrophobicity", "1": "RKEDQN", "2": "GASTPHY", "3": "CLVIMFW"}
# # # '1' -> Polar; '2' -> Neutral, '3' -> Hydrophobicity

# # normalized_VDWV = {"name": "normalized_VDWV", "1": "GASTPD", "2": "NVEQIL", "3": "MHKFRYW"}
# # # '1' -> (0-2.78); '2' -> (2.95-4.0), '3' -> (4.03-8.08)

# # polarity = {"name": "polarity", "1": "LIFWCMVY", "2": "CPNVEQIL", "3": "KMHFRYW"}
# # # '1' -> (4.9-6.2); '2' -> (8.0-9.2), '3' -> (10.4-13.0)

# # charge = {"name": "charge", "1": "KR", "2": "ANCQGHILMFPSTWYV", "3": "DE"}
# # # '1' -> Positive; '2' -> Neutral, '3' -> Negative

# # sec_struct = {"name": "secondary_struct", "1": "EALMQKRH", "2": "VIYCWFT", "3": "GNPSD"}
# # # '1' -> Helix; '2' -> Strand, '3' -> coil

# # solvent_accessibility = {"name": "solvent_accessibility", "1": "ALFCGIVW", "2": "RKQEND", "3": "MPSTHY"}
# # # '1' -> Buried; '2' -> Exposed, '3' -> Intermediate

# # polarizability = {"name": "polarizability", "1": "GASDT", "2": "CPNVEQIL", "3": "KMHFRYW"}
# # # '1' -> (0-0.108); '2' -> (0.128-0.186), '3' -> (0.219-0.409)



# # #1.)
# #         self.assertFalse(ctd.empty, 'Descriptor dataframe should not be empty')
# #         self.assertEqual(ctd.shape, (self.num_seqs[dataset], 147), 'Descriptor not of correct ({},147)'.format(self.num_seqs[dataset]))
# #         self.assertIsInstance(ctd, pd.DataFrame)
# #         self.assertTrue(ctd.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

# # #2.)
# #         #get descriptor values
# #         comp = desc.get_composition()
# # #2.)
# #         self.assertFalse(comp.empty, 'Descriptor dataframe should not be empty')
# #         self.assertEqual(comp.shape, (self.num_seqs[dataset], 3), 'Descriptor not of correct ({},3)'.format(self.num_seqs[dataset]))
# #         self.assertIsInstance(comp, pd.DataFrame)
# #         self.assertTrue(comp.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

# #         #get descriptor values
# #         trans = desc.get_transition()
# # #3.)
# #         self.assertFalse(trans.empty, 'Descriptor dataframe should not be empty')
# #         self.assertEqual(trans.shape, (self.num_seqs[dataset], 3), 'Descriptor not of correct ({},3)'.format(self.num_seqs[dataset]))
# #         self.assertIsInstance(trans, pd.DataFrame)
# #         self.assertTrue(trans.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

# #         #get descriptor values
# #         distr = desc.get_distribution()
# # #4.)
# #         self.assertFalse(distr.empty, 'Descriptor dataframe should not be empty')
# #         self.assertEqual(distr.shape, (self.num_seqs[dataset],15), 'Descriptor not of correct ({},15)'.format(self.num_seqs[dataset]))
# #         self.assertIsInstance(distr, pd.DataFrame)
# #         self.assertTrue(distr.any().isnull().sum()==0, 'Descriptor should not contain any null values.')




# from .autocorrelation import moreaubroto_autocorrelation, moran_autocorrelation, geary_autocorrelation
from .autocorrelation import *
from .composition import *
from .conjoint_triad import *
from .ctd import *
from .quasi_sequence_order import *

#protPy package metadata
__name__ = 'protpy'
__version__ = "0.0.2"
__description__ = "Python package for generating various biochemical, physiochemical and structural descriptors/features of protein sequences."
__author__ = 'AJ McKenna, https://github.com/amckenna41'
__authorEmail__ = 'amckenna41@qub.ac.uk'
__maintainer__ = "AJ McKenna"
__license__ = 'MIT'
__url__ = 'https://github.com/amckenna41/protPy'
__download_url__ = "https://github.com/amckenna41/protPy/archive/refs/heads/main.zip"
__status__ = "Production"
__keywords__ = ["bioinformatics", "protein engineering", "python", "pypi", "machine learning", \
                "aaindex", "protein descriptors", "physiochemical descriptors", "biochemical descriptors"
                "structural descriptors"]

#list of all available descriptors in protPy
all_descriptors = ["aa_composition", "dipeptide_composition", "tripeptide_composition",
    "moreaubroto_autocorrelation", "moran_autocorrelation", "geary_autocorrelation",
    "pseudo_amino_acid_composition", "amphiphilic_pseudo_amino_acid_composition", 
    "sequence_order_coupling_number", "conjoint_triad", "ctd_composition", 
    "ctd_transition", "ctd_distribution", "quasi_sequence_order"
    ]


###############################################################################
#####   Setup.py - installs all the required packages and dependancies    #####
###############################################################################

import pathlib
from setuptools import setup, find_packages
import sys
import protpy

#ensure python version is greater than 3
if (sys.version_info[0] < 3):
    sys.exit('Python 3 is the minimum version requirement.')

#get path to README file
HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

setup(name=protpy.__name__,
      version=protpy.__version__,
      description=protpy.__description__,
      long_description = README,
      long_description_content_type = "text/markdown",
      author=protpy.__license__,
      author_email=protpy.__authorEmail__,
      maintainer=protpy.__maintainer__,
      license=protpy.__license__,
      url=protpy.__url__,
      download_url=protpy.__download_url__,
      keywords=protpy.__keywords__,
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
      install_requires=[
        "aaindex",
        "numpy",
        "pandas",
        "scikit-learn",
        "tqdm",
        "varname"
      ],
     test_suite='tests',
     # packages=find_packages(), #create Manifest file to ignore results folder in dist
     packages=find_packages(),
     include_package_data=True,
     zip_safe=False)



###########################################################################
###############      protPy - Composition Module Tests      ###############
###########################################################################

import pandas as pd
import numpy as np
import os
import unittest
import re
import random
from Bio import SeqIO
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import protpy as protpy

class ProtPyCompositionTests(unittest.TestCase):
    """
    Test suite for testing composition module and functionality 
    in protpy package. 

    Test Cases
    ----------
    test_amino_acid_composition:
        testing correct protpy amino acid composition functionality.
    test_dipeptide_composition:
        testing correct protpy dipeptide amino acid composition functionality.
    test_tripeptide_composition:
        testing correct protpy tripeptide amino acid composition functionality.
    test_pseudo_amino_acid_composition:
        testing correct protpy pseudo amino acid composition functionality.
    test_amphiphilic_pseudo_amino_acid_composition:
        testing correct protpy amphophilic amino acid composition functionality.
    """
    def setUp(self):
        """ Import protein sequences from test fasta files using Biopython package. """
        #using next() to get first item (protein seq) from SeqIO Generator
        with open(os.path.join("tests", "test_fasta1.fasta")) as pro:
            self.protein_seq1 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta2.fasta")) as pro:
            self.protein_seq2 = str(next(SeqIO.parse(pro,'fasta')).seq)
        
        with open(os.path.join("tests", "test_fasta3.fasta")) as pro:
            self.protein_seq3 = str(next(SeqIO.parse(pro,'fasta')).seq)

        with open(os.path.join("tests", "test_fasta4.fasta")) as pro:
            self.protein_seq4 = str(next(SeqIO.parse(pro,'fasta')).seq)

        #list of canonical amino acids
        self.amino_acids = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", 
            "Q", "R", "S", "T", "V", "W", "Y"]

    def test_amino_acid_composition(self):
        """ Testing Amino Acid Composition protein descriptor attributes and methods. """
#1.)
        aa_composition_seq1 = protpy.aa_composition(self.protein_seq1)
        aa_composition_seq1_expected_values = pd.DataFrame(columns=["A", "C", "D", "E", "F", "G", "H", 
            "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"])
        aa_composition_seq1_expected_values.loc[0] = [6.693, 3.108, 5.817, 3.347, 6.614, 6.295, 1.195, 
            6.215, 4.781, 7.888, 1.594, 6.454, 4.542, 4.382, 3.108, 7.649, 7.888, 7.251, 0.876, 4.303]

        self.assertIsInstance(aa_composition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(aa_composition_seq1.shape, (1, 20), 'Descriptor not of correct shape.') 
        self.assertEqual(self.amino_acids, list(aa_composition_seq1.columns), 'Incorrect column values found.')
        self.assertTrue(aa_composition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(aa_composition_seq1.equals(aa_composition_seq1_expected_values))
#2.)
        aa_composition_seq2 = protpy.aa_composition(self.protein_seq2)
        aa_composition_seq2_expected_values = pd.DataFrame(columns=["A", "C", "D", "E", "F", "G", "H", 
            "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"])
        aa_composition_seq2_expected_values.loc[0] = [8.057, 0.0, 5.213, 3.318, 3.081, 10.664, 1.185, 
            2.607, 6.872, 6.161, 1.659, 5.924, 7.346, 8.057, 7.346, 8.294, 7.82, 2.607, 1.185, 2.607]

        self.assertIsInstance(aa_composition_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(aa_composition_seq2.shape, (1, 20), 'Descriptor not of correct shape.') 
        self.assertEqual(self.amino_acids, list(aa_composition_seq2.columns), 'Incorrect column values found.')
        self.assertTrue(aa_composition_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(aa_composition_seq2.equals(aa_composition_seq2_expected_values))
#3.)
        aa_composition_seq3 = protpy.aa_composition(self.protein_seq3)
        aa_composition_seq3_expected_values = pd.DataFrame(columns=["A", "C", "D", "E", "F", "G", "H", 
            "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"])
        aa_composition_seq3_expected_values.loc[0] = [5.263, 3.947, 1.316, 3.947, 5.263, 2.632, 0.0, 
            3.947, 2.632, 18.421, 1.316, 6.579, 2.632, 0.0, 2.632, 9.211, 6.579, 18.421, 0.0, 5.263]

        self.assertIsInstance(aa_composition_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(aa_composition_seq3.shape, (1, 20), 'Descriptor not of correct shape.') 
        self.assertEqual(self.amino_acids, list(aa_composition_seq3.columns), 'Incorrect column values found.')
        self.assertTrue(aa_composition_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(aa_composition_seq3.equals(aa_composition_seq3_expected_values))
#4.)
        aa_composition_seq4 = protpy.aa_composition(self.protein_seq4)
        aa_composition_seq4_expected_values = pd.DataFrame(columns=["A", "C", "D", "E", "F", "G", "H", 
            "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"])
        aa_composition_seq4_expected_values.loc[0] = [8.831, 0.0, 5.728, 2.864, 3.103, 10.263, 0.955, 
            3.341, 7.399, 6.444, 1.671, 5.251, 6.683, 8.353, 6.921, 8.831, 7.637, 1.909, 1.193, 2.625]

        self.assertIsInstance(aa_composition_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(aa_composition_seq4.shape, (1, 20), 'Descriptor not of correct shape.') 
        self.assertEqual(self.amino_acids, list(aa_composition_seq4.columns), 'Incorrect column values found.')
        self.assertTrue(aa_composition_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(aa_composition_seq4.equals(aa_composition_seq4_expected_values))
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            aa_composition_seq5 = protpy.aa_composition(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            aa_composition_seq5 = protpy.aa_composition(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            aa_composition_seq5 = protpy.aa_composition(invalid_seq7)

    def test_dipeptide_composition(self):
        """ Testing Dipeptide Composition protein descriptor attributes and functionality. """
#1.)
        dipeptide_comp_seq1 = protpy.dipeptide_composition(self.protein_seq1)

        self.assertIsInstance(dipeptide_comp_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(dipeptide_comp_seq1.shape, (1, 400), 'Descriptor not of correct shape.') 
        for col in list(dipeptide_comp_seq1.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{2}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(dipeptide_comp_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.float64 for col in list(dipeptide_comp_seq1.dtypes)), "")
#2.)
        dipeptide_comp_seq2 = protpy.dipeptide_composition(self.protein_seq2)

        self.assertIsInstance(dipeptide_comp_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(dipeptide_comp_seq2.shape, (1, 400), 'Descriptor not of correct shape.') 
        for col in list(dipeptide_comp_seq2.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{2}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(dipeptide_comp_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.float64 for col in list(dipeptide_comp_seq2.dtypes)), "")
#3.)
        dipeptide_comp_seq3 = protpy.dipeptide_composition(self.protein_seq3)

        self.assertIsInstance(dipeptide_comp_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(dipeptide_comp_seq3.shape, (1, 400), 'Descriptor not of correct shape.') 
        for col in list(dipeptide_comp_seq3.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{2}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(dipeptide_comp_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.float64 for col in list(dipeptide_comp_seq3.dtypes)), "")
#4.)
        dipeptide_comp_seq4 = protpy.dipeptide_composition(self.protein_seq4)

        self.assertIsInstance(dipeptide_comp_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(dipeptide_comp_seq4.shape, (1, 400), 'Descriptor not of correct shape.') 
        for col in list(dipeptide_comp_seq4.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{2}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(dipeptide_comp_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.float64 for col in list(dipeptide_comp_seq4.dtypes)), "")
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            dipeptide_composition_seq5 = protpy.dipeptide_composition(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            dipeptide_composition_seq5 = protpy.dipeptide_composition(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            dipeptide_composition_seq5 = protpy.dipeptide_composition(invalid_seq7)

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping.")
    def test_tripeptide_composition(self):
        """ Testing Tripeptide Composition protein descriptor attributes and functionality. """
#1.)
        tripeptide_comp_seq1 = protpy.tripeptide_composition(self.protein_seq1)

        print("tripeptide_comp_seq1")
        print(tripeptide_comp_seq1.dtypes)
        self.assertIsInstance(tripeptide_comp_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(tripeptide_comp_seq1.shape, (1, 8000), 'Descriptor not of correct shape.') 
        for col in list(tripeptide_comp_seq1.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{3}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(tripeptide_comp_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.int64 for col in list(tripeptide_comp_seq1.dtypes)), "")
#2.)
        tripeptide_comp_seq2 = protpy.tripeptide_composition(self.protein_seq2)

        self.assertIsInstance(tripeptide_comp_seq2, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(tripeptide_comp_seq2.shape, (1, 8000), 'Descriptor not of correct shape.') 
        for col in list(tripeptide_comp_seq2.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{3}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(tripeptide_comp_seq2.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.int64 for col in list(tripeptide_comp_seq2.dtypes)), "")
#3.)
        tripeptide_comp_seq3 = protpy.tripeptide_composition(self.protein_seq3)

        self.assertIsInstance(tripeptide_comp_seq3, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(tripeptide_comp_seq3.shape, (1, 8000), 'Descriptor not of correct shape.') 
        for col in list(tripeptide_comp_seq3.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{3}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(tripeptide_comp_seq3.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.int64 for col in list(tripeptide_comp_seq3.dtypes)), "")
#4.)
        tripeptide_comp_seq4 = protpy.tripeptide_composition(self.protein_seq4)

        self.assertIsInstance(tripeptide_comp_seq4, pd.DataFrame, 'Descriptor not of type DataFrame.')
        self.assertEqual(tripeptide_comp_seq4.shape, (1, 8000), 'Descriptor not of correct shape.') 
        for col in list(tripeptide_comp_seq4.columns):
            #check all columns follow pattern of XY where x & y are amino acids 
            self.assertTrue(bool(re.match(r'^[A-Z]{3}$', col)), "")      
            self.assertIn(col[0], self.amino_acids, "")
            self.assertIn(col[1], self.amino_acids, "")
        self.assertTrue(tripeptide_comp_seq4.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
        self.assertTrue(all(col == np.int64 for col in list(tripeptide_comp_seq4.dtypes)), "")
#5.)
        invalid_seq5 = "ABCDEF"
        with (self.assertRaises(ValueError)):
            tripeptide_composition_seq5 = protpy.tripeptide_composition(invalid_seq5)
#6.)
        invalid_seq6 = "OOOOO"
        with (self.assertRaises(ValueError)):
            tripeptide_composition_seq5 = protpy.tripeptide_composition(invalid_seq6)
#7.)
        invalid_seq7 = 12345
        with (self.assertRaises(TypeError)):
            tripeptide_composition_seq5 = protpy.tripeptide_composition(invalid_seq7)

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping.")
    def test_pseudo_amino_acid_composition(self):
        """ Testing pseudo aa composition protein descriptor attributes and functionality. """
        lambda_range = range(30, 60, 10) 
        weight_range = []
        for i in range(10): weight_range.append(random.random())
#1.)
        for lam in lambda_range:
            for weight in weight_range:
                pseudo_amino_acid_composition_seq1 = protpy.pseudo_amino_acid_composition(self.protein_seq1,
                    weight=weight, lamda=lam)
            
                self.assertIsInstance(pseudo_amino_acid_composition_seq1, pd.DataFrame, 'Descriptor not of type DataFrame.')
                self.assertEqual(pseudo_amino_acid_composition_seq1.shape, (1, 50), 'Descriptor not of correct shape.') 
                # for col in list(pseudo_amino_acid_composition_seq1.columns):
                #     #check all columns follow pattern of XY where x & y are amino acids 
                #     self.assertTrue(bool(re.match(r'^[A-Z]{3}$', col)), "")      
                self.assertTrue(pseudo_amino_acid_composition_seq1.any().isnull().sum()==0, 'Descriptor should not contain any null values.')
                self.assertTrue(all(col == np.float64 for col in list(pseudo_amino_acid_composition_seq1.dtypes)), "")

# def pseudo_amino_acid_composition(sequence, lamda=30, weight=0.05, properties=["ARGP820101", "KUHL950101"]):

    def amp_pseudo_amino_acid_composition(self):
        """ """
        #random list of floats for weight from 0 - 1
        #random lambda values from 30 - 100
        pass
# def amphiphilic_pseudo_aac(sequence, lamda=30, weight=0.5, properties=["ARGP820101", "KUHL950101"]):

